import collections

class State(object):
    ''' State Machine Class.
        Initializes to start_state, and optionally an iterator of states, or
        an iterator of (start_state, end_state) transitions.

        Errors: ValueError is raised if a transition is not allowed.
    '''
    def __init__(self, start_state, transitions=None):
        self.current_state = start_state  # Initialize a starting state.
        self.transitions = collections.defaultdict(list)

        # Initialize transitions if passed.
        if transitions is not None:
            for start, check, action, end in transitions:
                self.add_transition(start, check, action, end)

    @property
    def states(self):
        return self.transitions.keys()

    def add_transition(self, start, check, action, end):
        '''Adds an allowed state transition.
        start::String
        check:: pointer to a method with no parameters which returns a bool
            or a property of an object.
        action:: pointer to a method with no parameters
        end::String
        '''
        self.transitions[start].append((check, action, end))

    def run(self):
        current_state = self.current_state
        possible_transitions = self.transitions[current_state]
        # Perform checks until a valid transition is found.
        # Perform the action, then change the state.
        for check, action, end in possible_transitions:
            if hasattr(check, "__call__"):
                should_continue = check.__call__()
            else:
                should_continue = check
            if should_continue:
                action.__call__()
                self.current_state = end
        # Raise StateTransitionError if no valid checks.
        else:
            raise


class StateError(Exception):
    pass



import json
'''
This library implements the basic logic for the Pengine classes.
    - PengineBuilder: An interface for Pengine objects.
    - PengineState: For the current state of a Pengine Query, used in Pengine
    class.
    - Pengine: Base class used for performing queries -- this is the access
    point.
'''


class Query(object):
    '''
    Query objects are the objects which hold hte current query information.

    pengine :: Pengine The pengine this query is bound to.
    ask:: String describing the query.
    queryMaster :: Bool Immediately run query
    '''
    def __init__(self, pengine, ask, queryMaster=True):
        '''pengine :: Pengine, ask :: String, queryMaster :: Boolean'''
        if pengine.debug:
            print("Initializing query with query: {}".format(ask))
        self.availProofs = []
        self.hasMore = True
        self.pengine = pengine
        self.ask = ask
        if queryMaster:
            pengine.doAsk(self, ask)
        return None

    def dumpDebugState(self):
        pass

    def hasNext(self):
        '''Return True if we __think__ we have more data, else False.'''
        return self.hasMore or not self.availProofs == []

    def __iter__(self):
        return self
    
    def __next__(self):
        p = self.pengine
        if self.availProofs != []:
            data = self.availProofs.pop(0)
            if not self.hasMore and self.availProofs == []:
                p.iAmFinished(self)

            return data
        else:
            if self.hasMore:
                p.doNext(p.currentQuery)
                if self.availProofs != []:
                    return self.availProofs.pop(0)
            raise StopIteration

    def noMore(self):
        if not self.hasMore:
            return False

        self.hasMore = False
        if not self.availProofs:
            self.pengine.iAmFinished(self)

    def stop(self):
        ''' Stop the query on the slave.  We're done.
        Equivalent to typing period at the top-level in Prolog.

        Raises: PengineNotReadyException
        '''
        self.pengine.doStop()
        self.hasMore = False
        self.availProofs = []
        self.pengine.iAmFinished(self)

    def addNewData(self, data):
        '''
        Returns True if we __think__ we have more data.
        data::String
        '''
        for item in data:
            self.availProofs.append(item)
        return self.hasMore or not self.availProofs == []



from pengines.Exceptions import CouldNotCreateException, \
    PengineNotReadyException, \
    PengineNotAvailableException
from pengines.State import State
from pengines.Builder import PengineBuilder
from pengines.Query import Query
import copy
from urllib.request import Request, urlopen
import json
import urllib


class Pengine(object):
    def __init__(self, builder=None, slave_limit=-1, debug=False):
        '''
        This Pengine object is used to run one or more queries of the
        Prolog Knowledge base set in the PengineBuilder.  builder::None ||
        PengineBuilder  Uses a new PengineBuilder with default settings,
        otherwise use the supplied PengineBuilder initialized with the
        desired settings.  A deepcopy is performed to preserve state.
        slave_limit::Int Sets limits on the number of slaves that this
        pengine can have.
        '''
        self.availOutput = None
        self.currentQuery = None
        self.state = State("not_created")

        # Initialize debug value, defaulting to False.
        self.debug = debug

        # Handle the builder logic.
        if builder is None:
            self.po = PengineBuilder()
        else:
            self.po = copy.deepcopy(builder)

        # self.create() gets pengineID
        self.slave_limit = slave_limit
        try:
            self.pengineID = self.create()
        except PengineNotReadyException:
            self.state.current_state = "destroy"
            raise PengineNotReadyException("Pengine could not be created!!")

        if self.debug:
            print("Initialization complete.")
        return None

    def ask(self, query):
        '''
        Parameters:
            query::str -> The prolog query as a string.

        Return:
            query::pengine.Query.Query() object.

        Errors:
            IOError raised if query cannot be created, or server cannot be
            contacted
        '''
        if self.debug:
            print("Starting the call to ask.")
            print("Current state is {}".format(self.state.current_state))
        self.currentQuery = Query(self, query, False)
        print("Call to ask is complete")
        return self.currentQuery

    def doAsk(self, query):
        # Check to make sure state is idle and can handle queries.
        if self.state.current_state != "idle":
            raise PengineNotReadyException("Not in a state to handle queries")
        # Begin running query.
        if self.currentQuery is None:
            self.currentQuery = query

        # Set pengine state to "ask", process response.
        self.state.current_state = "ask"
        answer = self.penginePost(self.po.getActualURL("send", self.pengineID),
                                  "application/x-prolog; charset=UTF-8",
                                  self.po.getRequestBodyAsk(query.ask,
                                                            self.getID()))
        self.handleAnswer(answer)

    def create(self):
        '''
        Configures the Pengine object.  Returns a pengine id string.

        Modifies state -- to "idle" if created successfully.  Else "destroy"
        '''
        assert self.state.current_state == "not_created"
        if self.debug:
            print("Starting call.")
        # Post the create request.
        url = self.po.getActualURL("create")
        contentType = "application/json"
        body = self.po.getRequestBodyCreate()
        if self.debug:
            print("Starting post request with URL {0}, content_type: {1}"\
                  ", and body: {2}".format(url, contentType, body))
        response = self.penginePost(url, contentType, body)
        if self.debug:
            print(response)

        # Begin setting various attributes.

        # Set the event state to destroy if request failed, or idle and waiting
        # for a query.
        event_string = response["event"]
        if event_string == "destroy":
            self.state.current_state = "destroy"
        elif event_string == "create":
            self.state.current_state = "idle"
        else:
            err_msg = "Create request event was {}, but must be create or"\
                "destroy".format(event_string)
            raise CouldNotCreateException(err_msg)

        # Handle setting ask.
        if self.po.ask is not None:
            # If a query is present, immediately handle it.
            self.currentQuery = Query(self, self.po.ask, False)
            self.state.current_state = "ask"

        # Handle "answer" key
        if "answer" in response:
            self.handleAnswer(response["answer"])

        # Handle "id" key
        id_ = response["id"]
        if id_ is None:
            raise CouldNotCreateException("No pengine id in create message")
        return id_

    def destroy(self):
        pass

    def doNext(self, query):
        '''
        This method asks the Pengine object to request the next proof from the
        server.

        Parameters:
            query :: Query the query object to continue providing data to.
        '''
        if self.debug:
            print("pengines.Pengine().doNext is firing with query ", query)
        # ToDo assert self.state.current_state == "ask"
        if query != self.currentQuery:
            raise PengineNotReadyException("Cannot advance more than one query -\
                                           finish one before starting next.")

        url = self.po.getActualURL("send", self.getID())
        contentType = "application/x-prolog; charset=UTF-8"
        body = self.po.getRequestBodyNext()
        if self.debug:
            print("url: {0}; body: {2}; contentType: {1}".format(url,
                                                            contentType,
                                                            body))
        string_response_object = self.penginePost(url, contentType, body)
        self.handleAnswer(string_response_object)

    def doPullResponse(self):
        '''
        Returns a string off the server, if it has one to give.
            Otherwise, returns None.

        @raises PengineNotReadyExceptioni
        '''
        # Check it is in the correct state.
        if self.state.current_state != "idle" and \
                self.state.current_state != "ask":
            return None

        # Initialize Post request
        url = self.po.getActualURL("pull_response", self.getID())
        content_type = "application/x-prolog; charset=UTF-8"
        body_response = self.po.getRequestBodyPullResponse()
        response = self.penginePost(url, content_type, body_response)
        # Pass the response the handler
        self.handleAnswer(response)
        raise TypeError("doPullNot implemented exception")

    def doStop(self):
        '''Raises PengineNotReadyException'''
        assert self.state.current_state == "ask"
        respObject = self.penginePost(self.po.getActualURL("send",
                                                           self.getID()),
                                      "application/x-prolog; charset=UTF-8",
                                      self.po.getRequestBodyStop())
        self.handleAnswer(respObject)
        pass

    def dumpStateDebug(self):
        pass

    def iAmFinished(self, query):
        '''
        Query won't use pengine again.

        query:: Query() : The query we are through with
        '''
        if self.currentQuery == query:
            self.currentQuery = None

        if self.state.current_state == "ask":
            self.state.current_state = "idle"

    def penginePost(self, url, contentType, body):
        '''
        Posts body to the Pengine.
        url:: String of url
        contentType :: String -> Value of Content-Type header
        body :: String -> Body of POST request

        Return: response_dict::dict -> A dict representing the JSON encoded response.
        Errors: IOError -> If there is a problem posting, an IOError is raised.
        '''
        if self.debug:
            print("Starting Post request.")
        # Set up request header
        header = dict()
        header["User-Agent"] = "PythonPengine"
        header["Accept"] = "application/json"
        header["Accept-Language"] = "en-us,en;q=0.5"
        header["Content-type"] = contentType

        # Make sure body is utf-8
        if isinstance(body, bytes):
            body_utf8 = body
        elif isinstance(body, str):
            body_utf8 = body.encode("utf-8")
        else:
            raise TypeError("Don't know how to handle body parameter of type\
                            {}").format(type(body))

        if self.debug:
            print("URL is: ", url)
            print("Data (body) : ", body_utf8)
            print("Headers: ", header)
        try:
            # Send Post Request -- catch errors and close
            request_object = Request(url, data=body_utf8, headers=header)
            response = urlopen(request_object)
            response_string_utf8 = response.read()
            response_string = response_string_utf8.decode("utf-8")
            if self.debug:
                print("response_string is :", response_string)
            response_dict = json.JSONDecoder().decode(response_string)
            # Catch bad status codes
            status = response.status
            if status < 200 or status > 299:
                err_msg = "Bad response code: {}  If query 500 was invalid?\
                    query threw Prolog exception".format(status)
                raise IOError(err_msg)
            return response_dict
        except IOError as e:
            self.destroy()
            raise e

    def handleAnswer(self, answer):
        '''
        answer::json The JSON response from the initial post to the server

        Exceptions:
            json.JSONDecodeError
            SyntaxError
        '''
        if self.debug:
            print("pengines.Pengine().handleAnswer({})".format(answer))
        try:
            if "event" in answer:
                event_val = answer["event"]
                if self.debug:
                    print("pengine.Pengine().handleAnswer reports {}".format(event_val))
                # Handle "success" switch
                if event_val == "success":
                    if "data" in answer:
                        # answer["data"] should be a list
                        self.currentQuery.addNewData(answer["data"])
                    if "more" in answer:
                        self.currentQuery.hasMore = answer["more"]

                # Handle "destroy" switch
                elif event_val == "destroy":
                    if "data" in answer:
                        self.handleAnswer(answer["data"])
                    if self.currentQuery is not None:
                        self.currentQuery.noMore()
                        self.state.current_state = "destroyed"

                # Handle "failure" switch
                elif event_val == "failure":
                    self.currentQuery.noMore()

                # Handle "stop" switch
                elif event_val == "stop":
                    self.currentQuery.noMore()

                # Handle "error" switch
                elif event_val == "error":
                    raise SyntaxError("Error - probably invalid Prolog query?")

                # Handle "died" switch
                elif event_val == "died":
                    self.state.current_state = "destroyed"

                # Default to raising a syntax error.
                else:
                    raise SyntaxError("Bad event in answer")
        except PengineNotReadyException:
            raise SyntaxError

    def getID(self):
        '''
        Getter method for the getID attribute.
        @return :: pengineID String
        @raises :: Assertion Error if current state not "ask" or "idle"
        '''
        current_state = self.state.current_state
        assert current_state == "ask" or current_state == "idle"
        return self.pengineID



''' This module handles all of the exceptions specific to the Python
Pengines logic.'''

# CouldNotCreateException
class CouldNotCreateException(Exception):
    def __init__(self, message):
        self.message = message

# PengineNotReadyException
class PengineNotReadyException(Exception):
    def __init__(self, message):
        self.message = message

# PengineNotAvailableException
class PengineNotAvailableException(Exception):
    def __init__(self, message):
        self.message = message



import json
import sys
import logging
from pengines.Exceptions import PengineNotReadyException


class PengineBuilder(object):
    def __init__(self,
                 urlserver=None,
                 application=None,
                 ask=None,
                 chunk=100,
                 destroy=True,
                 srctext=None,
                 srcurl=None,
                 format_type="json",
                 alias=None,
                 debug=False):
        self.alias = alias
        self.application = application
        self.ask = ask
        self.chunk = chunk
        self.destroy = destroy
        self.format_type = format_type
        self.urlserver = urlserver
        self.srctext = srctext
        self.srcurl = srcurl
        self.request_body = self.getRequestBodyCreate()
        self.debug = False

    def getRequestBodyCreate(self):
        '''
        String a representation of the request body for the create action.
        Return:
            json encoding of self.destroy, self.chunk, self.format_type,
            self.srctext, self.srcurl
        '''
        data = dict()
        if not self.destroy:
            data["destroy"] = False

        if self.chunk > 1:
            data["chunk"] = self.chunk

        data["format"] = self.format_type

        if self.srctext is not None:
            data["src_text"] = self.srctext

        if self.srcurl is not None:
            data["srcurl"] = self.srcurl

        if self.ask is not None:
            data["ask"] = self.ask
            
        if self.application is not None:
            data["application"] = self.application

        return json.JSONEncoder().encode(data)

    def getRequestBodyAsk(self, ask, id=None):
        '''
        ask::String The prolog query.
        id::String The id of the pengine id that is transmitting.
            Currently not used.
        '''
        return "ask({},[]).".format(ask)

    def dumpDebugState(self):
        '''Dumps debug information to stderr'''
        # Initialize the destroy string printout to stderr
        if self.destroy:
            destroy_string = "destroy at end of query"
        else:
            destroy_string = "retain at end of query"

        serialized = ["--- PengineBuilder ----",
                      "alias {}".format(self.alias),
                      "application {}".format(self.application),
                      "ask {}".format(self.ask),
                      "chunk size {}".format(self.chunk),
                      destroy_string,
                      "server {}".format(self.urlserver),
                      "srctext {}".format(self.srctext),
                      "srcurl {}".format(self.srcurl),
                      "--- end PengineBuilder ---"]
        for line in serialized:
            sys.stderr.write(line)

    def getActualURL(self, action, id_=None):
        '''
        This class dispatches to self._getActualURL(action, id_) or
            self._getActualURL_(action) depending on the presence of id.
        returns the full url needed to accomplish the task.
        '''
        if id_ is not None:
            return self._getActualURL(action, id_)
        else:
            return self._getActualURL_(action)

    def _getActualURL_(self, action):
        '''
        action :: String -> an action like "create", "destroy", etc.
        return :: String -> The full URL needed to perform the action.
        '''
        if self.urlserver is None:
            raise PengineNotReadyException("Need to set server to get URL")
        # Verify the server url ends in a '/'
        if self.urlserver[-1] != "/":
            urlserver = self.urlserver + "/"
        else:
            urlserver = self.urlserver
        relative = "pengine/{}".format(action)
        new = urlserver + relative
        logging.debug("New value is : {}".format(new))
        return new

    def _getActualURL(self, action, id_):
        '''
        Parses in the relative information necessary to perform a query on
        Pengines into a full URL form.

        action::String -> Action to be performed "send", "post", etc.
        id::String -> The id of the pengine in question.
        '''
        # Verify server url is set, otherwise raise PengineNotReadyException
        print("_getActualURL is firing.")
        if self.urlserver is None:
            raise PengineNotReadyException("Cannot get actual URL without \
                                           setting the server")
        # Catch missing ending '/' if applicable.
        if self.urlserver[-1] != "/":
            urlserver = self.urlserver + "/"
        else:
            urlserver = self.urlserver

        relative = "pengine/{0}?format=json&id={1}".format(action, id_)
        url = urlserver + relative
        print("_getActualURL is returning {}".format(url))
        return url

    def getRequestBodyNext(self):
        '''
        Returns the POST body to get the next result.
        Todo: Remove this Javaism.
        '''
        return "next."

    def getReqestBodyStop(self):
        '''
        Returns 'stop
        Todo: Remove this Javaism.
        '''
        return "stop."



from setuptools import setup
setup(
    name = "pengines",
    packages = ["pengines"],
    version = "0.1.8",
    description = "A simple python library for interacting with SWI-Prologs Pengines",
    long_description=open("README.rst").read(),
    author = "Ian Andrich, Chris Mungall",
    author_email = "iandrich87+github@gmail.com",
    url = "https://github.com/ian-andrich/PythonPengines",
    download_url = "https://github.com/ian-andrich/PythonPengines/0.1.4",

    keywords = ["pengines", "pengine", "prolog", "swipl", "swi-prolog", "logic programming"],
    classifiers = [],
)
